from scipy.stats import pearsonr,spearmanr
from scipy.spatial.distance import euclidean
import math
import numpy as np

def spearmanDistance(u,v):
    if len(set(u)) == 1 and len(set(v)) == 1 and u[0] == v[0]: #Prevent NaN values
        return 0
    elif len(set(u)) == 1 and len(set(v)) == 1 and u[0] != v[0]:
        return 1
    elif len(set(u)) == 1 or len(set(v)) == 1:
        return euclidean(u,v)/math.sqrt(len(v)) #normalized euclidean distance
    return 1 - spearmanr(u,v)[0]

def pearsonDistance(u,v):
    if len(set(u)) == 1 and len(set(v)) == 1 and u[0] == v[0]:  # Prevent NaN values
        return 0
    elif len(set(u)) == 1 and len(set(v)) == 1 and u[0] != v[0]:
        return 1
    elif len(set(u)) == 1 or len(set(v)) == 1:
        return euclidean(u, v) / math.sqrt(len(v))  # normalized euclidean distance
    return 1 - pearsonr(u,v)[0]

def find_elbow(data_list): #data is a list of y axis numbers. Assume x is 1 to len(data)
    '''
    Inspired by: https://datascience.stackexchange.com/questions/57122/in-elbow-curve-how-to-find-the-point-from-where-the-curve-starts-to-rise
    '''

    data = np.array(list(zip(range(1,len(data_list)+1),data_list)))
    theta = get_data_theta(data)

    # make rotation matrix
    co = np.cos(theta)
    si = np.sin(theta)
    rotation_matrix = np.array(((co, si), (-si, co)))

    # rotate data vector
    rotated_vector = data.dot(rotation_matrix)

    # return x value of elbow
    return np.where(rotated_vector[:,1] == rotated_vector[:,1].min())[0][0] + 1 #+1 transforms index to actual x value

def get_data_theta(data): #in radians
    return np.arctan2(abs(data[:, 1].max() - data[:, 1].min()), abs(data[:, 0].max() - data[:, 0].min()))



'''
Methods related to Hierarchical Clustering
'''
import numpy as np
from fastcluster import linkage
import copy
import random

########CLUSTER TREE METHODS############################################################################################
def createClusterTree(linkage_matrix,instance_labels,instances):
    '''
    :param linkage_matrix: array of [[cluster_index_1,cluster_index_2,distance_between_clusters,# observations in new cluster],...]
    :param instance_labels: array of instance labels, in order given to the the mechanism that created the above linkage matrix
    :param instances: 2D numpy.ndarray of instances, where rows are the object to be clustered.
    :return: ClusterTree based on linkage matrix and instance_labels
    '''

    return ClusterTree(linkage_matrix,instance_labels,instances)

class ClusterTree:
    def __init__(self,linkage_matrix,instance_labels,instances):
        self.clusters = []
        self.instances = instances
        self.label_to_index_map = {}
        self.instance_labels = instance_labels
        self.significant_clusters = None
        self.significant_cluster_objs = None

        #Map instance labels to instance indices
        for index in range(len(instance_labels)):
            self.label_to_index_map[instance_labels[index]] = index

        #Populate initial instances
        for label in instance_labels:
            self.clusters.append(Cluster([self.label_to_index_map[label]],None,None,None,0))

        #Traverse linkage matrix to create tree
        cluster_rank = 1
        for row in linkage_matrix:
            cluster_index_1 = int(round(row[0]))
            cluster_index_2 = int(round(row[1]))
            distance_between_clusters = row[2]
            combined_instance_indices = self.clusters[cluster_index_1].instance_index_array+self.clusters[cluster_index_2].instance_index_array
            new_cluster = Cluster(combined_instance_indices,cluster_index_1,cluster_index_2,distance_between_clusters,cluster_rank)
            self.clusters.append(new_cluster)

            #Set sibling and parent indices
            self.clusters[cluster_index_1].parent_index = len(self.clusters) - 1
            self.clusters[cluster_index_2].parent_index = len(self.clusters) - 1
            self.clusters[cluster_index_1].sibling_index = cluster_index_2
            self.clusters[cluster_index_2].sibling_index = cluster_index_1

            cluster_rank += 1

    def rootCluster(self):
        return self.clusters[-1]

    def getChildrenOf(self,cluster):
        return (self.clusters[cluster.child_index_1],self.clusters[cluster.child_index_2])

    def twoMeansClusterIndexOf(self,cluster):
        subcluster1,subcluster2 = self.getChildrenOf(cluster)
        mean_1 = self.getMeanInstanceOfClusterInstances(subcluster1)
        mean_2 = self.getMeanInstanceOfClusterInstances(subcluster2)
        mean_total = self.getMeanInstanceOfClusterInstances(cluster)

        ss1 = 0
        for instance_index in subcluster1.instance_index_array:
            instance = self.instances[instance_index]
            ss1 += pow(np.linalg.norm(mean_1-instance),2)

        ss2 = 0
        for instance_index in subcluster2.instance_index_array:
            instance = self.instances[instance_index]
            ss2 += pow(np.linalg.norm(mean_2-instance),2)

        tts = 0
        for instance_index in cluster.instance_index_array:
            instance = self.instances[instance_index]
            tts += pow(np.linalg.norm(mean_total-instance),2)

        return (ss1+ss2)/tts

    def getMeanInstanceOfClusterInstances(self,cluster):
        instance_indices = cluster.instance_index_array
        num_attributes = self.instances[0].size
        num_instances = len(instance_indices)
        mean_array = np.empty(num_attributes)
        for index in instance_indices:
            mean_array += self.instances[index]
        return mean_array/num_instances

    def getSDInstanceOfClusterInstances(self,cluster):
        instance_indices = cluster.instance_index_array
        array = []
        for index in instance_indices:
            array.append(self.instances[index])
        array = np.transpose(np.array(array))
        std_array = []
        for a in array:
            std_array.append(np.std(a))
        return np.array(std_array)

    def getRandomInstances(self,mean_array,sd_array,num_samples):
        total = []
        for i in range(mean_array.size):
            a = np.random.normal(mean_array[i],sd_array[i],num_samples)
            total.append(a)
        return np.transpose(np.array(total))

    def indicesToLabels(self,array):
        l = []
        for i in array:
            b = []
            for a in i:
                b.append(self.instance_labels[a])
            l.append(b)
        return l

    def getNSignificantClusters(self,N,metric,method,p_value=0.05,sample_count=30,random_state=None):
        if self.significant_clusters == None:
            self.getSignificantClusters(metric,method,p_value,sample_count,random_state=random_state)

        copy_objs = copy.deepcopy(self.significant_cluster_objs)
        if N > len(copy_objs):
            N = len(copy_objs)

        while len(copy_objs) > N:
            #Find lowest rank cluster that also has a sibling
            min_rank = np.inf
            lowest_cluster = None
            for cluster in copy_objs:
                if cluster.height_tracker < min_rank:
                    #Find sibling, if it exists.
                    c_index = self.clusters[cluster.sibling_index].sibling_index
                    potential_sibling_cluster = None
                    for c in copy_objs:
                        if c.sibling_index == c_index:
                            potential_sibling_cluster = cluster
                    if potential_sibling_cluster != None:
                        min_rank = cluster.height_tracker
                        lowest_cluster = cluster

            #Get lowest rank cluster's sibling
            own_index = self.clusters[lowest_cluster.sibling_index].sibling_index
            sibling_cluster = None
            for cluster in copy_objs:
                if cluster.sibling_index == own_index:
                    sibling_cluster = cluster

            #Find lowest rank cluster's parent
            parent_cluster = copy.deepcopy(self.clusters[lowest_cluster.parent_index])

            #Remove siblings from copy. Add parent
            copy_objs.remove(lowest_cluster)
            copy_objs.remove(sibling_cluster)
            copy_objs.append(parent_cluster)

        #Convert copy into list of lists of labels
        sig_clusters = []
        colors = []
        for c in copy_objs:
            sig_clusters.append(c.instance_index_array)
            colors.append(c.hex)
        return self.indicesToLabels(sig_clusters),colors

    def getSignificantClusters(self,metric,method,p_value=0.05,sample_count=30,random_state=None):
        if random_state != None:
            random.seed(random_state)
            np.random.seed(random_state)

        if self.significant_clusters == None:
            self.significant_clusters = []
            self.significant_cluster_objs = []
            root_cluster = self.rootCluster()
            self.getSignificantClusters_Recur(root_cluster,p_value,sample_count,metric,method)

            colors = []
            for c in self.significant_cluster_objs:
                colors.append(c.hex)
            return self.indicesToLabels(self.significant_clusters),colors
        else:
            colors = []
            for c in self.significant_cluster_objs:
                colors.append(c.hex)
            return self.indicesToLabels(self.significant_clusters),colors

    def getSignificantClusters_Recur(self,cluster,p_value,sample_count,metric,method):
        if cluster.child_index_1 != None:
            #observed_CI = self.twoMeansClusterIndexOf(cluster)
            observed_CI = cluster.distance_between_children
            cluster_P_value = self.getPValue(cluster,sample_count,observed_CI,metric,method)
            significance_cutoff = p_value*(cluster.getNumInstances()-1)/(len(self.instances)-1)
            if cluster_P_value <= significance_cutoff:
                subcluster1,subcluster2 = self.getChildrenOf(cluster)
                self.getSignificantClusters_Recur(subcluster1,p_value,sample_count,metric,method)
                self.getSignificantClusters_Recur(subcluster2,p_value,sample_count,metric,method)
            elif cluster_P_value > significance_cutoff:
                self.significant_clusters.append(cluster.instance_index_array)
                self.significant_cluster_objs.append(cluster)
        else:
            self.significant_clusters.append(cluster.instance_index_array)
            self.significant_cluster_objs.append(cluster)

    def getPValue(self,cluster,sample_count,observed_CI,metric,method):
        num_below_observed = 0

        cluster_mean = self.getMeanInstanceOfClusterInstances(cluster)
        cluster_sd = self.getSDInstanceOfClusterInstances(cluster)

        for i in range(sample_count):
            #randomly sample instances based off of cluster instance mean and SD
            array = self.getRandomInstances(cluster_mean,cluster_sd,len(cluster.instance_index_array))
            labels = list(range(len(cluster.instance_index_array)))
            linkage_matrix = linkage(array,metric=metric,method=method)

            #Create Cluster Tree from linkage matrix and get CI of root cluster
            sub_tree = ClusterTree(linkage_matrix,labels,array)
            #sub_tree_ci = sub_tree.twoMeansClusterIndexOf(sub_tree.rootCluster())
            sub_tree_ci = sub_tree.rootCluster().distance_between_children

            #If CI of root cluster is less/greater than or equal to observed, increment
            #if sub_tree_ci <= observed_CI: #For 2 Means computation
            if sub_tree_ci >= observed_CI: #For linkage distance computation
                num_below_observed += 1
        return num_below_observed/sample_count

class Cluster:
    def __init__(self,instance_index_array,child_index_1,child_index_2,distance_between_chidren,height_tracker):
        '''
        :param instance_index_array: list of instance indices that make up this cluster
        :param child_index_1/2: index of children in ClusterTree's clusters attribute
        :param distance_between_chidren: linkage distance between children
        :param height_tracker: an integer that ranks the cluster's height on the tree. Lower rank => closer to leaves
        '''
        self.instance_index_array = instance_index_array
        self.child_index_1 = child_index_1
        self.child_index_2 = child_index_2
        self.distance_between_children = distance_between_chidren
        self.height_tracker = height_tracker
        self.parent_index = None
        self.sibling_index = None
        self.hex = self.randomHex()

    def getNumInstances(self):
        return len(self.instance_index_array)

    def randomHex(self):
        s = '#'
        for i in range(6):
            s += random.choice(['0', '1', '2', '3', '4', '5', '6', '7', '8', '9', 'A', 'B', 'C', 'D', 'E', 'F'])
        return s




import sys
import time
import pickle
import seaborn
import csv
import matplotlib.pyplot as plt
from scipy.spatial.distance import cdist, pdist, squareform
import os
import math
import numpy as np
import pandas as pd
from LCSDIVE import HClust
from sklearn.metrics import silhouette_score
import matplotlib as mpl
from LCSDIVE.Utilities import find_elbow

def job(experiment_path, rule_height_factor):
    job_start_time = time.time()

    # Load information
    file = open(experiment_path + '/phase1pickle', 'rb')
    phase1_pickle = pickle.load(file)
    file.close()

    cv_count = phase1_pickle[9]
    full_info = phase1_pickle[1]

    data_headers = full_info[2]
    class_label = phase1_pickle[8]
    random_state = phase1_pickle[7]

    cv_info = phase1_pickle[0]

    cmap = mpl.colors.ListedColormap(['xkcd:navy blue','xkcd:yellow'])

    # Merge Rule Population
    merged_population = []
    cv_starting_indices = []
    cv_index_counter = 0
    for cv in range(cv_count):
        cv_starting_indices.append(cv_index_counter)
        file = open(experiment_path + '/CV_' + str(cv) + '/model', 'rb')
        model = pickle.load(file)
        for rule in model.population.popSet:
            merged_population.append(rule)
            cv_index_counter += 1

    num_rules = len(merged_population)
    rule_specificity_array = []
    micro_to_macro_rule_index_map = {}
    micro_rule_index_count = 0
    macro_rule_index_count = 0

    current_cv_counter = 0
    cv_headers = None
    cv_attribute_map = None # maps the cv's feature name to index in universal data headers
    for inst in range(num_rules):
        if inst in cv_starting_indices:
            cv_headers = cv_info[current_cv_counter][10]
            cv_attribute_map = {}
            for feature_name in cv_headers:
                cv_attribute_map[feature_name] = np.where(data_headers == feature_name)[0][0]
            current_cv_counter += 1

        rule = merged_population[inst]
        a = [0]*len(data_headers)
        for cv_attribute_index in rule.specifiedAttList:
            feature_name = cv_headers[cv_attribute_index]
            feature_index = cv_attribute_map[feature_name]
            a[feature_index] = 1

        for microclassifier in range(rule.numerosity):
            rule_specificity_array.append(a)
            micro_to_macro_rule_index_map[micro_rule_index_count] = macro_rule_index_count
            micro_rule_index_count += 1
        macro_rule_index_count += 1
    rule_specificity_array = np.array(rule_specificity_array)

    rule_df = pd.DataFrame(rule_specificity_array, columns=data_headers, index=list(range(micro_rule_index_count)))

    plt.figure(figsize=((10 / math.sqrt(rule_height_factor), 10 * math.sqrt(rule_height_factor))))
    h = seaborn.heatmap(rule_df, cmap=cmap, cbar_kws={'ticks':[0,1]})
    h.tick_params(left=False,labelleft=False)
    if rule_specificity_array.shape[1] <= 11:
        h.set_xticklabels(h.get_xmajorticklabels(),fontsize='xx-large')
    elif rule_specificity_array.shape[1] <= 20:
        h.set_xticklabels(h.get_xmajorticklabels(),fontsize='x-large')
    if rule_specificity_array.shape[1] >= 20:
        plt.xticks(rotation=90)
    plt.xlabel('Features',fontsize='xx-large')
    plt.ylabel('Rules',fontsize='xx-large')
    plt.tight_layout()
    plt.savefig(experiment_path + '/Composite/rulepop/rulepopHeatmap.png')
    plt.close('all')

    # Rule Population Clustermaps
    r = seaborn.clustermap(rule_df, metric='sqeuclidean', method='ward', cmap=cmap, cbar_kws={'ticks':[0,1]})

    feature_order_indices = r.dendrogram_col.reordered_ind
    new_feature_order = []
    for i in feature_order_indices:
        new_feature_order.append(data_headers[i])
    with open(experiment_path + '/Composite/rulepop/clustered_feature_order_left_to_right.csv', mode='w') as file:
        writer = csv.writer(file, delimiter=',', quotechar='"', quoting=csv.QUOTE_MINIMAL)
        writer.writerow(new_feature_order)
    file.close()

    rule_cluster_tree = HClust.createClusterTree(r.dendrogram_row.linkage, list(range(micro_rule_index_count)),rule_df.to_numpy())
    rule_clusters, rule_colors = rule_cluster_tree.getSignificantClusters(p_value=0.05, sample_count=100,metric='sqeuclidean', method='ward',random_state=random_state)

    rule_distortions = []

    precomputed_distances = squareform(pdist(rule_specificity_array, metric='sqeuclidean'))
    silhouettes = []

    for rule_cluster_count in reversed(range(1, len(rule_clusters) + 1)):
        if not os.path.exists(
                experiment_path + '/Composite/rulepop/ruleclusters/' + str(rule_cluster_count) + '_clusters'):
            os.mkdir(experiment_path + '/Composite/rulepop/ruleclusters/' + str(rule_cluster_count) + '_clusters')

        rule_subclusters, rule_colors = rule_cluster_tree.getNSignificantClusters(rule_cluster_count, p_value=0.05,sample_count=100, metric='sqeuclidean',method='ward',random_state=random_state)

        # Elbow Method
        centroids = []
        for cluster in rule_subclusters:
            centroid = np.zeros(len(data_headers))
            for inst_label_index in cluster:
                centroid += rule_specificity_array[inst_label_index]
            centroid /= len(cluster)
            centroids.append(centroid)
        centroids = np.array(centroids)
        rule_distortions.append(sum(np.min(cdist(rule_specificity_array, centroids, 'sqeuclidean'), axis=1)))

        # Silhouette Method
        s_counter = 0
        new_l = [0] * micro_rule_index_count
        for cluster in rule_subclusters:
            for inst_label_index in cluster:
                new_l[inst_label_index] = s_counter
            s_counter += 1
        if rule_cluster_count != 1:
            silhouettes.append(silhouette_score(precomputed_distances, new_l, metric='precomputed'))
        else:
            silhouettes.append(0)

        # Clustermaps
        rule_color_dict = {}
        rule_color_count = 0
        for cluster in rule_subclusters:
            random_color = rule_colors[rule_color_count]
            for inst_label in cluster:
                rule_color_dict[inst_label] = random_color
            rule_color_count += 1
        rule_color_list = pd.Series(dict(sorted(rule_color_dict.items())))
        rule_color_list = pd.Series.to_frame(rule_color_list)
        rule_color_list.columns = ['Found']

        c = seaborn.clustermap(rule_df, row_linkage=r.dendrogram_row.linkage, col_linkage=r.dendrogram_col.linkage,row_colors=rule_color_list, cmap=cmap, cbar_kws={'ticks':[0,1]}, figsize=(10 / math.sqrt(rule_height_factor), 10 * math.sqrt(rule_height_factor)))
        c.ax_heatmap.tick_params(right=False,labelright=False)
        if rule_specificity_array.shape[1] <= 11:
            c.ax_heatmap.set_xticklabels(c.ax_heatmap.get_xmajorticklabels(), fontsize='xx-large')
        elif rule_specificity_array.shape[1] <= 20:
            c.ax_heatmap.set_xticklabels(c.ax_heatmap.get_xmajorticklabels(),fontsize='x-large')
        c.ax_heatmap.set_xlabel('Features',fontsize='xx-large')
        c.ax_heatmap.set_ylabel('Rules',fontsize='xx-large',rotation=-90,labelpad=20)
        if rule_specificity_array.shape[1] >= 20:
            plt.setp(c.ax_heatmap.get_xticklabels(),rotation=90)
        plt.tight_layout()
        plt.savefig(experiment_path + '/Composite/rulepop/ruleclusters/' + str(rule_cluster_count) + '_clusters/ruleClustermap.png', dpi=300)
        plt.close('all')

        with open(experiment_path + '/Composite/rulepop/ruleclusters/' + str(rule_cluster_count) + '_clusters/ruleClusters.csv', mode='w') as file:
            writer = csv.writer(file, delimiter=',', quotechar='"', quoting=csv.QUOTE_MINIMAL)
            for rule_cluster in rule_subclusters:
                exp_color = rule_color_dict[rule_cluster[0]]
                writer.writerow(['ClusterID: ' + exp_color])
                writer.writerow(list(data_headers) + [class_label, 'Accuracy', 'Numerosity', 'Specificity', 'Init Timestamp'])

                spec_sum = np.array([0.0] * len(data_headers))
                acc_spec_sum = np.array([0.0] * len(data_headers))
                acc_sum = 0
                numerosity_sum = 0
                init_ts_sum = 0
                specificity_sum = 0

                covered_macro_rule_indices = []
                for inst_index in rule_cluster:
                    macro_rule_index = micro_to_macro_rule_index_map[inst_index]
                    if not macro_rule_index in covered_macro_rule_indices:
                        covered_macro_rule_indices.append(macro_rule_index)
                        rule = merged_population[macro_rule_index]

                        cv_range_indices = cv_starting_indices+[np.inf]
                        for start_i in range(len(cv_range_indices)-1):
                            start_val = cv_range_indices[start_i]
                            end_val = cv_range_indices[start_i+1]

                            if macro_rule_index >= start_val and macro_rule_index < end_val:
                                break

                        cv_headers = cv_info[start_i][10]
                        cv_attribute_map = {}
                        for feature_name in cv_headers:
                            cv_attribute_map[feature_name] = np.where(data_headers == feature_name)[0][0]


                        condition = [np.nan]*len(data_headers)
                        condition_counter = 0
                        for feature_i in range(len(cv_headers)):
                            if feature_i in rule.specifiedAttList:
                                feature_name = cv_headers[feature_i]
                                feature_index = cv_attribute_map[feature_name]
                                condition[feature_index] = rule.condition[condition_counter]
                                spec_sum[feature_index] += 1
                                acc_spec_sum[feature_index] += rule.accuracy
                                condition_counter += 1
                            else:
                                feature_name = cv_headers[feature_i]
                                feature_index = cv_attribute_map[feature_name]
                                condition[feature_index] = '#'

                        for cv_attribute_index in rule.specifiedAttList:
                            feature_name = cv_headers[cv_attribute_index]
                            feature_index = cv_attribute_map[feature_name]
                            condition[feature_index] = 1

                        writer.writerow(condition + [rule.phenotype, rule.accuracy, rule.numerosity,
                                                     len(rule.specifiedAttList) / len(data_headers),
                                                     rule.initTimeStamp])

                        acc_sum += rule.accuracy * rule.numerosity
                        numerosity_sum += rule.numerosity
                        init_ts_sum += rule.initTimeStamp * rule.numerosity
                        specificity_sum += len(rule.specifiedAttList) / len(data_headers) * rule.numerosity

                writer.writerow(['Rule Specificity Sums'])
                ks = []
                vs = []
                for k, v in sorted(dict(zip(list(data_headers), list(spec_sum))).items(), key=lambda item: item[1]):
                    ks.append(k)
                    vs.append(v)
                writer.writerow(list(reversed(ks)))
                writer.writerow(list(reversed(vs)))

                writer.writerow(['Rule Accuracy Weighted Specificity Sums'])
                ks = []
                vs = []
                for k, v in sorted(dict(zip(list(data_headers), list(acc_spec_sum))).items(), key=lambda item: item[1]):
                    ks.append(k)
                    vs.append(v)
                writer.writerow(list(reversed(ks)))
                writer.writerow(list(reversed(vs)))

                writer.writerow(['Avg Accuracy', 'Avg Init Timestamp', 'Avg Specificity'])
                writer.writerow([acc_sum / numerosity_sum, init_ts_sum / numerosity_sum, specificity_sum / numerosity_sum])

                writer.writerow([])
        file.close()

    # Plot Rule Elbow Plot
    rule_distortions.reverse()
    optimal_elbow = find_elbow(rule_distortions)
    plt.plot(range(1, len(rule_clusters) + 1), rule_distortions, 'bx-')
    plt.xlabel('Number of Clusters',fontsize='xx-large')
    plt.ylabel('Distortion',fontsize='xx-large')
    plt.axvline(x=optimal_elbow, color='xkcd:sky', linestyle='--',label='Elbow at: '+str(optimal_elbow))
    plt.legend(fontsize='xx-large')
    plt.savefig(experiment_path + '/Composite/rulepop/' + str(optimal_elbow) + 'optimalClusters.png',dpi=300)
    plt.close('all')

    # Plot AT Silhouette Plot
    silhouettes.reverse()
    optimal_silhouette = np.argmax(np.array(silhouettes))+1
    plt.plot(range(1, len(rule_clusters) + 1), silhouettes, 'bx-')
    plt.xlabel('Number of Clusters', fontsize='xx-large')
    plt.ylabel('Silhouette Score', fontsize='xx-large')
    plt.axvline(x=optimal_silhouette, color='xkcd:sky', linestyle='--', label='Max at: ' + str(optimal_silhouette))
    plt.legend(fontsize='xx-large')
    plt.savefig(experiment_path + '/Composite/rulepop/' + str(optimal_silhouette) + 'optimalSilhouette.png', dpi=300)
    plt.close('all')

    # Save Runtime
    runtime_file = open(experiment_path + '/Composite/rulepop/runtime.txt', 'w')
    runtime_file.write(str(time.time() - job_start_time))
    runtime_file.close()

    # Print completion
    print("Rule phase 2 complete")

if __name__ == '__main__':
    job(sys.argv[1],float(sys.argv[2]))


import sys
import argparse
import pickle
import pygame
from pygame.locals import *
import math
from scipy.spatial import distance
import copy

'''Sample Run Code
python NetworkVisualization.py --o ../Outputs/Final --e dietadj
'''

def main(argv):
    # Parse arguments
    parser = argparse.ArgumentParser(description="")
    parser.add_argument('--o', dest='output_path', type=str, help='path to output directory')
    parser.add_argument('--e', dest='experiment_name', type=str, help='name of experiment (no spaces)')
    parser.add_argument('--nodep', dest='node_power', type=int, default=3)
    parser.add_argument('--edgep', dest='edge_power', type=int, default=3)
    parser.add_argument('--nodes', dest='node_maxsize', type=int, default=50) #radius
    parser.add_argument('--edges', dest='edge_maxsize', type=int, default=30)
    parser.add_argument('--labelshow', dest='show_labels', type=int, default=0)
    parser.add_argument('--labels', dest='label_maxsize', type=int, default=50)
    parser.add_argument('--fromsave', dest='from_save', type=int, default=1)
    options = parser.parse_args(argv[1:])

    output_path = options.output_path
    experiment_name = options.experiment_name
    experiment_path = output_path + '/' + experiment_name
    edge_power = options.edge_power
    node_power = options.node_power
    edge_maxsize = options.edge_maxsize
    node_maxsize = options.node_maxsize
    show_labels = options.show_labels == 1
    label_maxsize = options.label_maxsize
    from_save = options.from_save == 1

    # Extract Network Data
    save_opened = False
    if from_save:
        try:
            file = open(experiment_path + '/Composite/rulepop/savedpygame', 'rb')
            info = pickle.load(file)
            file.close()
            save_opened = True
        except:
            file = open(experiment_path + '/Composite/rulepop/networkpickle', 'rb')
            info = pickle.load(file)
            file.close()
    else:
        file = open(experiment_path + '/Composite/rulepop/networkpickle', 'rb')
        info = pickle.load(file)
        file.close()

    acc_spec_dict = info[0]
    edge_list = info[1]
    weight_list = info[2]
    pos = info[3]

    # Pygame
    pygame.init()
    main_clock = pygame.time.Clock()
    window_width = 1000
    window_height = 1000
    window_surface = pygame.display.set_mode((window_width, window_height))
    BLACK = (0, 0, 0)
    WHITE = (255,255,255)
    NODECOLOR = (255,51,119)
    EDGECOLOR = (224,184,255)

    originalasd = copy.deepcopy(acc_spec_dict)
    originalel = copy.deepcopy(edge_list)
    originalwl = copy.deepcopy(weight_list)

    #Normalize Sizes
    max_node_value = max(acc_spec_dict.values())
    for i in acc_spec_dict:
        acc_spec_dict[i] = math.pow(acc_spec_dict[i] / max_node_value, node_power) * node_maxsize  # Cubic Node Size Function

    max_weight_value = max(weight_list)
    for i in range(len(weight_list)):
        weight_list[i] = math.pow(weight_list[i] / max_weight_value, edge_power) * edge_maxsize  # Cubic Weight Function

    # Set up nodes' initial positions:
    nodes = {}
    for nodename in pos:
        x = int(int(pos[nodename][0] * window_width * 0.4) + window_width / 2)
        y = int(int(pos[nodename][1] * window_height * 0.4) * -1 + window_height / 2)
        nodes[nodename] = [acc_spec_dict[nodename],x,y]

    # Set up initial labels
    if not save_opened:
        label_display = {}
        for nodename in acc_spec_dict:
            font = pygame.font.SysFont("monospace", max(20,int(acc_spec_dict[nodename]/node_maxsize*label_maxsize)))
            label_display[nodename] = [show_labels,font]
    else:
        label_display = {}
        for nodename in acc_spec_dict:
            font = pygame.font.SysFont("monospace", max(20,int(acc_spec_dict[nodename] / node_maxsize * label_maxsize))) #font minsize is 10
            label_display[nodename] = [info[4][nodename], font]

    dragged_nodes = []

    while True:
        for event in pygame.event.get():
            if event.type == QUIT:
                newpos = {}
                labels = {}
                for nodename in nodes:
                    tx = (nodes[nodename][1]-window_width/2)/(window_width*0.4)
                    ty = (nodes[nodename][2]-window_height/2)/(window_height*0.4)*-1
                    newpos[nodename] = [tx,ty]
                    labels[nodename] = label_display[nodename][0]
                to_save = [originalasd,originalel,originalwl,newpos,labels]
                outfile = open(experiment_path + '/Composite/rulepop/savedpygame', 'wb')
                pickle.dump(to_save, outfile)
                outfile.close()

                pygame.image.save(window_surface,experiment_path + '/Composite/rulepop/saved.png')

                pygame.quit()
                sys.exit()
            if event.type == KEYDOWN:
                if event.key == ord('l'):
                    mx,my = pygame.mouse.get_pos()
                    for nodename in nodes:
                        dist = distance.euclidean((nodes[nodename][1],nodes[nodename][2]),(mx,my))
                        if dist < nodes[nodename][0]:
                            label_display[nodename][0] = not label_display[nodename][0]

            if event.type == pygame.MOUSEBUTTONDOWN:
                mx, my = pygame.mouse.get_pos()
                for nodename in nodes:
                    dist = distance.euclidean((nodes[nodename][1], nodes[nodename][2]), (mx, my))
                    if dist < nodes[nodename][0]:
                        if not nodename in dragged_nodes:
                            dragged_nodes.append(nodename)

            if event.type == pygame.MOUSEBUTTONUP:
                dragged_nodes = []

        window_surface.fill(WHITE)

        mx, my = pygame.mouse.get_pos()
        for nodename in dragged_nodes:
            nodes[nodename][1] = mx
            nodes[nodename][2] = my

        edgecounter = 0
        for n1,n2 in edge_list:
            weight = weight_list[edgecounter]
            pygame.draw.line(window_surface,EDGECOLOR,(nodes[n1][1],nodes[n1][2]),(nodes[n2][1],nodes[n2][2]),int(weight))
            edgecounter += 1

        for nodename in nodes:
            x = nodes[nodename][1]
            y = nodes[nodename][2]
            pygame.draw.circle(window_surface,NODECOLOR,(x,y),int(nodes[nodename][0]))
            if label_display[nodename][0]:
                label = label_display[nodename][1].render(nodename,1,BLACK)
                size = label_display[nodename][1].size(nodename)
                window_surface.blit(label,(x-size[0]/2,y-size[1]/2))

        pygame.display.update()
        main_clock.tick()

if __name__ == '__main__':
    sys.exit(main(sys.argv))


import pickle
import networkx as nx
import matplotlib.pyplot as plt
import math
import sys
import numpy as np

def job(experiment_path):
    file = open(experiment_path + '/phase1pickle', 'rb')
    phase1_pickle = pickle.load(file)
    file.close()

    full_info = phase1_pickle[1]
    data_headers = full_info[2]
    cv_count = phase1_pickle[9]
    cv_info = phase1_pickle[0]

    models = []
    for cv in range(cv_count):
        file = open(experiment_path + '/CV_' + str(cv) + '/model', 'rb')
        model = pickle.load(file)
        models.append(model)

    # Rule Specificity Network
    attribute_acc_specificity_counts = np.zeros(len(data_headers))
    merged_attribute_cooccurrences = []
    cv_counter = 0
    for model in models:
        cv_headers = cv_info[cv_counter][10]
        cv_attribute_map = {}
        for feature_name in cv_headers:
            cv_attribute_map[feature_name] = np.where(data_headers == feature_name)[0][0]

        cv_spec_list = model.get_final_attribute_specificity_list()
        transformed_cv_spec_list = [0]*len(data_headers)
        for c in range(len(cv_headers)):
            feature_name = cv_headers[c]
            transformed_cv_spec_list[cv_attribute_map[feature_name]] = cv_spec_list[c]

        attribute_acc_specificity_counts += np.array(transformed_cv_spec_list)
        attribute_cooccurrences = model.get_final_attribute_coocurrences(cv_headers, len(cv_headers))
        if merged_attribute_cooccurrences == []:
            merged_attribute_cooccurrences = attribute_cooccurrences
        else:
            for index in range(len(attribute_cooccurrences)):
                pair = attribute_cooccurrences[index]
                shouldAdd = True
                for index2 in range(len(merged_attribute_cooccurrences)):
                    pair2 = merged_attribute_cooccurrences[index2]
                    if (pair[0] == pair2[0] and pair[1] == pair2[1]) or (pair[0] == pair2[1] and pair[1] == pair2[0]):
                        merged_attribute_cooccurrences[index2] += attribute_cooccurrences[index]
                        shouldAdd = False
                if shouldAdd:
                    merged_attribute_cooccurrences.append(attribute_cooccurrences[index])
        cv_counter += 1

    acc_spec_dict = {}
    for header_index in range(len(data_headers)):
        acc_spec_dict[data_headers[header_index]] = attribute_acc_specificity_counts[header_index]

    G = nx.Graph()
    edge_list = []
    weight_list = []
    for co in merged_attribute_cooccurrences:
        G.add_edge(co[0], co[1], weight=co[3])
        edge_list.append((co[0], co[1]))
        weight_list.append(co[3])

    pos = nx.spring_layout(G, k=1)

    to_save = [acc_spec_dict, edge_list, weight_list, pos]
    outfile = open(experiment_path + '/Composite/rulepop/networkpickle', 'wb')
    pickle.dump(to_save, outfile)
    outfile.close()

    max_node_value = max(acc_spec_dict.values())
    for i in acc_spec_dict:
        acc_spec_dict[i] = math.pow(acc_spec_dict[i] / max_node_value, 3) * 1000  # Cubic Node Size Function

    max_weight_value = max(weight_list)
    for i in range(len(weight_list)):
        weight_list[i] = math.pow(weight_list[i] / max_weight_value, 3) * 10  # Cubic Weight Function

    nx.draw_networkx_nodes(G, pos=pos, nodelist=acc_spec_dict.keys(), node_size=[v * 1 for v in acc_spec_dict.values()], node_color='#FF3377')
    nx.draw_networkx_edges(G, pos=pos, edge_color='#E0B8FF', edgelist=edge_list, width=[v * 1 for v in weight_list])
    nx.draw_networkx_labels(G, pos=pos)
    plt.axis('off')
    plt.savefig(experiment_path + '/Composite/rulepop/rulepopGraph.png', dpi=300)
    plt.close('all')

if __name__ == '__main__':
    job(sys.argv[1])


import sys
import os
import argparse
import time
from LCSDIVE import AnalysisPhase2ATJob
from LCSDIVE import AnalysisPhase2RuleJob
from LCSDIVE import AnalysisPhase2NetworkJob

'''Sample Run Code
python AnalysisPhase2.py --o ../Outputs --e mp6 --cluster 0
python AnalysisPhase2.py --o ../Outputs --e mp11 --cluster 0
python AnalysisPhase2.py --o ../Outputs --e mp20 --cluster 0

python AnalysisPhase2.py --o /Users/robert/Desktop/outputs/test1/mp6/viz-outputs --e root --cluster 0
python AnalysisPhase2.py --o /Users/robert/Desktop/outputs/test1/mp11/viz-outputs --e root --cluster 0
'''

def main(argv):
    # Parse arguments
    parser = argparse.ArgumentParser(description="")
    parser.add_argument('--o', dest='output_path', type=str, help='path to output directory')
    parser.add_argument('--e', dest='experiment_name', type=str, help='name of experiment (no spaces)')
    parser.add_argument('--rheight', dest='rule_height_factor', type=float, default=1)
    parser.add_argument('--aheight', dest='at_height_factor', type=float, default=1)
    parser.add_argument('--cluster', dest='do_cluster', type=int, default=1)
    parser.add_argument('--am1', dest='at_memory1', type=int, default=2)
    parser.add_argument('--am2', dest='at_memory2', type=int, default=3)
    parser.add_argument('--rm1', dest='rule_memory1', type=int, default=5)
    parser.add_argument('--rm2', dest='rule_memory2', type=int, default=6)
    parser.add_argument('--nm1', dest='network_memory1', type=int, default=2)
    parser.add_argument('--nm2', dest='network_memory2', type=int, default=3)
    parser.add_argument('--dorule', dest='do_rule', type=int, default=1)

    options = parser.parse_args(argv[1:])

    output_path = options.output_path
    experiment_name = options.experiment_name
    rule_height_factor = options.rule_height_factor
    at_height_factor = options.at_height_factor
    do_cluster = options.do_cluster
    at_memory1 = options.at_memory1
    at_memory2 = options.at_memory2
    rule_memory1 = options.rule_memory1
    rule_memory2 = options.rule_memory2
    network_memory1 = options.network_memory1
    network_memory2 = options.network_memory2
    do_rule_cluster = options.do_rule
    experiment_path = output_path + '/' + experiment_name

    # CV Composite Analysis
    if not os.path.exists(experiment_path + '/Composite'):
        os.mkdir(experiment_path + '/Composite')
        os.mkdir(experiment_path + '/Composite/rulepop')
        os.mkdir(experiment_path + '/Composite/rulepop/ruleclusters')
        os.mkdir(experiment_path + '/Composite/at')
        os.mkdir(experiment_path + '/Composite/at/atclusters')

    if do_cluster == 1:
        submitClusterATJob(experiment_path,at_height_factor,at_memory1,at_memory2)
        submitClusterNetworkJob(experiment_path,network_memory1,network_memory2)
        if do_rule_cluster == 1:
            submitClusterRuleJob(experiment_path,rule_height_factor,rule_memory1,rule_memory2)
    else:
        submitLocalATJob(experiment_path,at_height_factor)
        if do_rule_cluster == 1:
            submitLocalRuleJob(experiment_path,rule_height_factor)
        submitLocalNetworkJob(experiment_path)

    ####################################################################################################################

def submitLocalATJob(experiment_path,at_height_factor):
    AnalysisPhase2ATJob.job(experiment_path,at_height_factor)

def submitLocalRuleJob(experiment_path,rule_height_factor):
    AnalysisPhase2RuleJob.job(experiment_path,rule_height_factor)

def submitLocalNetworkJob(experiment_path):
    AnalysisPhase2NetworkJob.job(experiment_path)

def submitClusterATJob(experiment_path,at_height_factor,m1,m2):
    job_ref = str(time.time())
    job_name = experiment_path + '/jobs/' + job_ref + '_run.sh'
    sh_file = open(job_name, 'w')
    sh_file.write('#!/bin/bash\n')
    sh_file.write('#BSUB -J ' + job_ref + '\n')
    sh_file.write('#BSUB -o ' + experiment_path + '/logs/' + job_ref + '.o\n')
    sh_file.write('#BSUB -e ' + experiment_path + '/logs/' + job_ref + '.e\n')

    this_file_path = os.path.dirname(os.path.realpath(__file__))
    sh_file.write('python ' + this_file_path + '/AnalysisPhase2ATJob.py ' + experiment_path + ' '+ str(at_height_factor)+'\n')
    sh_file.close()
    os.system('bsub -q i2c2_normal -R "rusage[mem='+str(m1)+'G]" -M '+str(m2)+'G < ' + job_name)

def submitClusterRuleJob(experiment_path,rule_height_factor,m1,m2):
    job_ref = str(time.time())
    job_name = experiment_path + '/jobs/' + job_ref + '_run.sh'
    sh_file = open(job_name, 'w')
    sh_file.write('#!/bin/bash\n')
    sh_file.write('#BSUB -J ' + job_ref + '\n')
    sh_file.write('#BSUB -o ' + experiment_path + '/logs/' + job_ref + '.o\n')
    sh_file.write('#BSUB -e ' + experiment_path + '/logs/' + job_ref + '.e\n')

    this_file_path = os.path.dirname(os.path.realpath(__file__))
    sh_file.write('python ' + this_file_path + '/AnalysisPhase2RuleJob.py ' + experiment_path + " " + str(rule_height_factor)+'\n')
    sh_file.close()
    os.system('bsub -q i2c2_normal -R "rusage[mem='+str(m1)+'G]" -M '+str(m2)+'G < ' + job_name)

def submitClusterNetworkJob(experiment_path,m1,m2):
    job_ref = str(time.time())
    job_name = experiment_path + '/jobs/' + job_ref + '_run.sh'
    sh_file = open(job_name, 'w')
    sh_file.write('#!/bin/bash\n')
    sh_file.write('#BSUB -J ' + job_ref + '\n')
    sh_file.write('#BSUB -o ' + experiment_path + '/logs/' + job_ref + '.o\n')
    sh_file.write('#BSUB -e ' + experiment_path + '/logs/' + job_ref + '.e\n')

    this_file_path = os.path.dirname(os.path.realpath(__file__))
    sh_file.write('python ' + this_file_path + '/AnalysisPhase2NetworkJob.py ' + experiment_path + '\n')
    sh_file.close()
    os.system('bsub -q i2c2_normal -R "rusage[mem=' + str(m1) + 'G]" -M ' + str(m2) + 'G < ' + job_name)


if __name__ == '__main__':
    sys.exit(main(sys.argv))

